from setuptools import setup, find_packages

setup(
    name='ruspeechstress',
    version='0.12',
    packages=find_packages(),
    url="https://github.com/pmashkovtseva/ruspeechstres",
    license='MIT',
    author='Polina Mashkovtseva',
    author_email='p.mashkovtseva@gmail.com',
    description='package for extending word stress dictionaries based on speech analysis',
)
